import numpy as np
from scipy.ndimage import filters
from scipy.ndimage import gaussian_filter
from scipy import ndimage
import math
from matplotlib.colors import hsv_to_rgb

def boundary(ypix,xpix):
    ''' returns pixels of mask that are on the exterior of the mask '''
    ypix = np.expand_dims(ypix.flatten(),axis=1)
    xpix = np.expand_dims(xpix.flatten(),axis=1)
    npix = ypix.shape[0]
    idist = ((ypix - ypix.transpose())**2 + (xpix - xpix.transpose())**2)
    idist[np.arange(0,npix),np.arange(0,npix)] = 500
    nneigh = (idist==1).sum(axis=1) # number of neighbors of each point
    iext = (nneigh<4).flatten()
    return iext

def init_masks(parent):
    '''creates RGB masks using stat and puts them in M0 or M1 depending on
    whether or not iscell is True for a given ROI
    args:
        ops: mean_image, Vcorr
        stat: xpix,ypix,xext,yext
        iscell: vector with True if ROI is cell
        ops_plot: plotROI, view, color, randcols
    outputs:
        M0: ROIs that are True in iscell
        M1: ROIs that are False in iscell
    '''
    ops = parent.ops
    stat = parent.stat
    iscell = parent.iscell
    cols = parent.ops_plot[3]
    ncells = len(stat)
    Ly = ops['Ly']
    Lx = ops['Lx']
    H      = np.zeros((cols.shape[1],Ly,Lx), np.float32)
    Sroi  = np.zeros((2,Ly,Lx), np.float32)
    Sext   = np.zeros((2,Ly,Lx), np.float32)
    LamAll = np.zeros((Ly,Lx), np.float32)
    Lam    = np.zeros((2,Ly,Lx), np.float32)
    iROI   = -1 * np.ones((Ly,Lx), np.int32)
    iExt   = -1 * np.ones((Ly,Lx), np.int32)

    for n in range(ncells):
        ypix = stat[n]['ypix']
        if ypix is not None:
            xpix = stat[n]['xpix']
            yext = stat[n]['yext']
            xext = stat[n]['xext']
            lam = stat[n]['lam']
            lam = lam / lam.sum()
            i = int(1-iscell[n])
            iROI[ypix,xpix] = n
            iExt[yext,xext] = n
            Sroi[i,ypix,xpix] = 1
            Sext[i,yext,xext] = 1
            Lam[i,ypix,xpix]  = lam
            LamAll[ypix,xpix] = lam

    # create H from cols and iROI
    for c in range(cols.shape[1]):
        H[c,iROI>=0] = cols[iROI[iROI>=0],c]

    LamMean = LamAll[LamAll>1e-10].mean()
    parent.H = H
    parent.iROI = iROI
    parent.iExt = iExt
    parent.Sroi = Sroi
    parent.Sext = Sext
    parent.Lam  = Lam
    parent.LamMean = LamMean

    # create all mask options
    parent.RGB_all = np.zeros((2,cols.shape[1],4,Ly,Lx,3), np.float32)
    parent.Vback = np.zeros((3,Ly,Lx), np.float32)
    parent.RGBback = np.zeros((3,Ly,Lx,3), np.float32)

    for i in range(2):
        for c in range(cols.shape[1]):
            for k in range(4):
                H = parent.H[c,:,:]
                if k<3:
                    S = parent.Sroi[i,:,:]
                else:
                    S = parent.Sext[i,:,:]
                V = np.maximum(0, np.minimum(1, 0.75*Lam[i,:,:]/parent.LamMean))
                if k>0:
                    if k==1:
                        mimg = ops['meanImg']
                        mimg = mimg - gaussian_filter(filters.minimum_filter(mimg,50),10)
                        mimg = mimg / gaussian_filter(filters.maximum_filter(mimg,50),10)
                        S = np.minimum(1, V*1.5)
                        mimg1 = np.percentile(mimg,1)
                        mimg99 = np.percentile(mimg,99)
                        mimg = (mimg - mimg1) / (mimg99 - mimg1)
                    elif k==2:
                        mimg = ops['meanImg']
                        S = np.minimum(1, V*1.5)
                        mimg1 = np.percentile(mimg,1)
                        mimg99 = np.percentile(mimg,99)
                        mimg = (mimg - mimg1) / (mimg99 - mimg1)
                    else:
                        vcorr = ops['Vcorr']
                        mimg = np.zeros((ops['Ly'],ops['Lx']),np.float32)
                        mimg[ops['yrange'][0]:ops['yrange'][1],
                            ops['xrange'][0]:ops['xrange'][1]] = vcorr
                        mimg = (mimg - mimg.min()) / (mimg.max() - mimg.min())

                    parent.Vback[k-1,:,:] = mimg
                    V = mimg
                    if k==3:
                        V = np.minimum(1, V + S)
                H = np.expand_dims(H,axis=2)
                S = np.expand_dims(S,axis=2)
                V = np.expand_dims(V,axis=2)
                hsv = np.concatenate((H,S,V),axis=2)
                parent.RGB_all[i,c,k,:,:,:] = hsv_to_rgb(hsv)

    for k in range(3):
        H = np.zeros((Ly,Lx,1),np.float32)
        S = np.zeros((Ly,Lx,1),np.float32)
        V = np.expand_dims(parent.Vback[k,:,:],axis=2)
        hsv = np.concatenate((H,S,V),axis=2)
        parent.RGBback[k,:,:,:] = hsv_to_rgb(hsv)

def make_chosen(M, ypix, xpix):
    v = M[ypix,xpix,:].max(axis=1)
    M[ypix,xpix,:] = np.resize(np.tile(v, 3), (3,len(ypix))).transpose()
    return M

def draw_masks(parent): #ops, stat, ops_plot, iscell, ichosen):
    '''creates RGB masks using stat and puts them in M0 or M1 depending on
    whether or not iscell is True for a given ROI
    args:
        ops: mean_image, Vcorr
        stat: xpix,ypix
        iscell: vector with True if ROI is cell
        ops_plot: plotROI, view, color, randcols
    outputs:
        M0: ROIs that are True in iscell
        M1: ROIs that are False in iscell
    '''
    ncells  = parent.iscell.shape[0]
    plotROI = parent.ops_plot[0]
    view    = parent.ops_plot[1]
    color   = parent.ops_plot[2]
    if view>0 and plotROI==0:
        M0 = parent.RGBback[view-1,:,:,:]
        M1 = parent.RGBback[view-1,:,:,:]
    else:
        ichosen = parent.ichosen
        wplot   = int(1-parent.iscell[ichosen])
        M0 = np.array(parent.RGB_all[0,color,view,:,:,:])
        M1 = np.array(parent.RGB_all[1,color,view,:,:,:])

        ypix    = parent.stat[ichosen]['ypix'].flatten()
        xpix    = parent.stat[ichosen]['xpix'].flatten()
        if wplot==0:
            M0 = make_chosen(M0, ypix, xpix)
        else:
            M1 = make_chosen(M1, ypix, xpix)
    return M0,M1

def flip_cell(parent):
    n = parent.ichosen
    i = int(1-parent.iscell[n])
    i0 = 1-i
    # cell indic
    nin = parent.iROI==n
    next = parent.iExt==n
    lam0 = np.array(parent.Lam[i0,nin])
    parent.Lam[i,nin] = lam0
    parent.Lam[i0,nin] = 0
    parent.Sroi[i,nin] = 1
    parent.Sroi[i0,nin] = 0
    parent.Sext[i,next] = 1
    parent.Sext[i0,next] = 0

    for i in range(2):
        for c in range(parent.H.shape[0]):
            for k in range(3):
                if k<2:
                    H = parent.H[c,nin]
                    S = parent.Sroi[i,nin]
                else:
                    H = parent.H[c,next]
                    S = parent.Sext[i,next]
                if k==0:
                    V = np.maximum(0, np.minimum(1, 0.75*parent.Lam[i,nin]/parent.LamMean))
                elif k==1:
                    V = parent.Vback[k-1,nin]
                    S = np.maximum(0, np.minimum(1, 1.5*0.75*parent.Lam[i,nin]/parent.LamMean))
                else:
                    V = parent.Vback[k-1,next]
                    V = np.minimum(1, V + S)
                H = np.expand_dims(H,axis=1)
                S = np.expand_dims(S,axis=1)
                V = np.expand_dims(V,axis=1)
                hsv = np.concatenate((H,S,V),axis=1)
                if k<2:
                    parent.RGB_all[i,c,k,nin,:] = hsv_to_rgb(hsv)
                else:
                    parent.RGB_all[i,c,k,next,:] = hsv_to_rgb(hsv)


def ROI_index(ops, stat):
    '''matrix Ly x Lx where each pixel is an ROI index (-1 if no ROI present)'''
    ncells = len(stat)-1
    Ly = ops['Ly']
    Lx = ops['Lx']
    iROI = -1 * np.ones((Ly,Lx), dtype=np.int32)
    for n in range(ncells):
        ypix = stat[n]['ypix'][~stat[n]['overlap']]
        if ypix is not None:
            xpix = stat[n]['xpix'][~stat[n]['overlap']]
            iROI[ypix,xpix] = n
    return iROI

def make_colorbar():
    H = np.arange(0,100).astype(np.float32)
    H = H / (100*1.3)
    H = H + 0.1
    H = 1 - H
    H = np.expand_dims(H,axis=1)
    S = np.ones((100,1))
    V = np.ones((100,1))
    hsv = np.concatenate((H,S,V), axis=1)
    colormat = hsv_to_rgb(hsv)
    colormat = np.expand_dims(colormat, axis=0)
    colormat = np.tile(colormat,(20,1,1))
    return colormat



import numpy as np
import math

def fitMVGaus(y,x,lam,thres=2.5):
    ''' computes 2D gaussian fit to data and returns ellipse of radius thres standard deviations
    inputs: 
        y, x, lam, thres
            y,x: pixel locations
            lam: pixel weights
        thres: number of standard deviations at which to draw ellipse
    outputs: 
        mu, cov, ellipse, area
            mu: mean of gaussian fit
            cov: covariance of gaussian fit
            radii: half of major and minor axis lengths of elliptical fit
            ellipse: coordinates of elliptical fit
            area: area of ellipse
    '''
    # normalize pixel weights
    lam = lam / lam.sum() 
    # mean of gaussian
    yx = np.stack((y,x))

    mu  = (lam*yx).sum(axis=-1)
    yx = yx - np.expand_dims(mu, axis=1)

    yx = yx * lam**.5
    #yx  = np.concatenate((y*lam**0.5, x*lam**0.5),axis=0)
    cov = yx @ yx.transpose()

    # radii of major and minor axes
    radii,evec  = np.linalg.eig(cov)
    radii       = thres * np.real(radii)**.5
    # compute pts of ellipse
    npts = 100

    p = np.expand_dims(np.linspace(0, 2*math.pi, npts),axis=1)

    p = np.concatenate((np.cos(p), np.sin(p)),axis=1)

    ellipse = (p * radii) @ evec.transpose() + mu

    area = (radii[0] * radii[1])**0.5 * math.pi

    radii  = np.sort(radii)[::-1]

    return mu, cov, radii, ellipse, area


import numpy as np
import time
import matplotlib.pyplot as plt
from numpy import fft
from numpy import random as rnd
from joblib import Parallel, delayed
import multiprocessing
import time

eps0 = 1e-5;
sigL = 0.85 # smoothing width for up-sampling kernels, keep it between 0.5 and 1.0...
lpad = 3   # upsample from a square +/- lpad 
smoothSigma = 1.15 # smoothing constant
maskSlope   = 2. # slope of taper mask at the edges
    
# smoothing kernel
def kernelD(a, b):        
    dxs = np.reshape(a[0], (-1,1)) - np.reshape(b[0], (1,-1))
    dys = np.reshape(a[1], (-1,1)) - np.reshape(b[1], (1,-1))
    
    ds = np.square(dxs) + np.square(dys)
    K = np.exp(-ds/(2*np.square(sigL)))    
    return K
    
def mat_upsample(lpad, ops):    
    lar    = np.arange(-lpad, lpad+1)
    larUP  = np.arange(-lpad, lpad+.001, 1./ops['subpixel'])

    x, y   = np.meshgrid(lar, lar)   
    xU, yU = np.meshgrid(larUP, larUP)   

    Kx = kernelD((x,y),(x,y))
    Kx = np.linalg.inv(Kx)
    Kg = kernelD((x,y),(xU,yU))
    
    Kmat = np.dot(Kx, Kg)
    nup = larUP.shape[0]
    
    return Kmat, nup


def prepareMasks(refImg):    
    i0,Ly,Lx = refImg.shape
    
    x = np.arange(0, Lx)
    y = np.arange(0, Ly)
    x = np.abs(x - x.mean())
    y = np.abs(y - y.mean())

    xx, yy = np.meshgrid(x, y)

    mY = y.max() - 4.
    mX = x.max() - 4.
    
    maskY = 1./(1.+np.exp((yy-mY)/maskSlope))
    maskX = 1./(1.+np.exp((xx-mX)/maskSlope))
    maskMul = maskY * maskX
    maskOffset = refImg.mean() * (1. - maskMul);

    #plt.imshow(maskMul)
    #plt.show()
    #plt.imshow(maskOffset)
    #plt.show()
    
    hgx = np.exp(-np.square(xx/smoothSigma))
    hgy = np.exp(-np.square(yy/smoothSigma))
    hgg = hgy * hgx
    hgg = hgg/hgg.sum()

    fhg = np.real(fft.fft2(fft.ifftshift(hgg))); # smoothing filter in Fourier domain

    cfRefImg   = np.conj(fft.fft2(refImg));
    absRef     = np.absolute(cfRefImg);
    cfRefImg   = cfRefImg / (eps0 + absRef) * fhg;
    
    return maskMul, maskOffset, cfRefImg

def correlation_map(data, cfRefImg):
    X = fft.fft2(data)
    J = X / (eps0 + np.absolute(X)) 
    J = J * cfRefImg
    cc = np.real(fft.ifft2(J))
    cc = fft.fftshift(cc, axes=(1,2))
    return cc, X

def getXYup(cc, Ls, ops):
    (lcorr, lpad, Lyhalf, Lxhalf) = Ls
    
    nimg = cc.shape[0]
    cc0 = cc[:, (Lyhalf-lcorr):(Lyhalf+lcorr+1), (Lxhalf-lcorr):(Lxhalf+lcorr+1)]

    cc0 = np.reshape(cc0, (nimg, -1))
    ix  = np.argmax(cc0, axis = 1)

    ymax, xmax = np.unravel_index(ix, (2*lcorr+1,2*lcorr+1))
    mxpt = [ymax+Lyhalf-lcorr, xmax + Lxhalf-lcorr]
    
    ccmat = np.zeros((nimg, 2*lpad+1, 2*lpad+1))
    for j in range(0, nimg):
        ccmat[j,:,:] = cc[j, (mxpt[0][j] -lpad):(mxpt[0][j] +lpad+1), (mxpt[1][j] -lpad):(mxpt[1][j] +lpad+1)]

    ccmat = np.reshape(ccmat, (nimg,-1))

    Kmat, nup = mat_upsample(lpad, ops)

    ccb = np.dot(ccmat, Kmat)
    imax = np.argmax(ccb, axis=1)
    cmax = np.amax(ccb, axis=1)

    ymax, xmax = np.unravel_index(imax, (nup,nup))

    mdpt = np.floor(nup/2)
    
    ymax,xmax = (ymax-mdpt)/ops['subpixel'], (xmax-mdpt)/ops['subpixel']
    
    ymax, xmax = ymax + mxpt[0] - Lyhalf, xmax + mxpt[1] - Lxhalf

    return ymax, xmax

def shift_data(X, ymax,xmax):
    nimg, Ly, Lx = X.shape
    
    Ny = fft.ifftshift(np.arange(-np.fix(Ly/2), np.ceil(Ly/2)))
    Nx = fft.ifftshift(np.arange(-np.fix(Lx/2), np.ceil(Lx/2)))
    [Nx,Ny] = np.meshgrid(Nx,Ny)
    Nx = Nx / Lx
    Ny = Ny / Ly
    
    dph = Nx * np.reshape(xmax, (-1,1,1)) + Ny * np.reshape(ymax, (-1,1,1))
    
    Y = np.real(fft.ifft2(X * np.exp((2j * np.pi) * dph)))
    return Y
    
def phasecorr_worker(data, refImg, ops):   
    
    nimg, Ly, Lx = data.shape
    refImg = np.reshape(refImg, (1, Ly, Lx))
    
    Lyhalf = int(np.floor(Ly/2))
    Lxhalf = int(np.floor(Lx/2))
    
    tstart = time.time()
    
    if ops['maxregshift']>0:
        maxregshift = ops['maxregshift'] 
    else:
        maxregshift = np.round(.1*np.maximum(Ly, Lx))

    lcorr = int(np.minimum(maxregshift, np.floor(np.minimum(Ly,Lx)/2.)-lpad))

    (maskMul, maskOffset, cfRefImg) = prepareMasks(refImg)

    data = data * maskMul + maskOffset

    cc, X = correlation_map(data, cfRefImg)

    ymax, xmax = getXYup(cc, (lcorr,lpad, Lyhalf, Lxhalf), ops)
    
    Y = shift_data(X, ymax,xmax)
    
    return Y, ymax, xmax, tstart

def phasecorr(data, refImg, ops, pool):    
    nimg = data.shape[0]
    
    num_cores = 20
    nbatch = int(np.ceil(nimg/float(num_cores)))
    
    inputs = range(0, nimg, nbatch)
    irange = []
    for i in inputs:
        irange.append(i + np.arange(0,np.minimum(nbatch, nimg-i)))

    t0 = time.time()
    
    results = [pool.apply_async(phasecorr_worker(data[irange[j],:, :], refImg, ops) for j in range(0,len(irange)))]
    #results = Parallel(n_jobs=num_cores)(delayed(phasecorr_worker)(data[irange[j],:, :], refImg, ops) for j in range(0,len(irange)))

    Y = np.zeros_like(data)
    ymax = np.zeros((nimg,))
    xmax = np.zeros((nimg,))
    tout = np.zeros((len(results),))
    for i in range(0,len(results)):
        Y[irange[i], :, :] = results[i][0]
        ymax[irange[i]] = results[i][1]
        xmax[irange[i]] = results[i][2]
        tout[i] = results[i][3] - t0

    return Y, ymax, xmax, tout




from PyQt5 import QtGui, QtCore
import pyqtgraph as pg
from pyqtgraph import console
import sys
import numpy as np
import os
import pickle
from suite2p import fig
from suite2p import run_s2p

### custom QDialog which allows user to fill in ops
# and run suite2p!
class RunWindow(QtGui.QDialog):
    def __init__(self, parent=None):
        super(RunWindow, self).__init__(parent)
        self.setGeometry(50,50,1100,700)
        self.setWindowTitle('Choose run options')
        self.win = QtGui.QWidget(self)
        self.layout = QtGui.QGridLayout()
        #layout = QtGui.QFormLayout()
        self.win.setLayout(self.layout)
        # initial ops values
        self.ops = run_s2p.default_ops()
        self.data_path = []
        self.save_path = []
        tifkeys = ['nplanes','nchannels','functional_chan','diameter','tau','fs']
        parkeys = ['num_workers','num_workers_roi']
        regkeys = ['nimg_init', 'batch_size', 'maxregshift', 'align_by_chan']
        cellkeys = ['navg_frames_svd','nsvd_for_roi','threshold_scaling']
        neukeys = ['ratio_neuropil_to_cell','inner_neuropil_radius','outer_neuropil_radius','min_neuropil_pixels']
        deconvkeys = ['win_baseline','sig_baseline','prctile_baseline','neucoeff']
        keys = [[],tifkeys, parkeys, regkeys, cellkeys, neukeys, deconvkeys]
        labels = ['Filepaths','Main settings','Parallel','Registration','Cell detection','Neuropil','Deconvolution']
        tooltips = ['each tiff has this many planes in sequence',
                    'each tiff has this many channels per plane',
                    'this channel is used to extract functional ROIs (1-based)',
                    'approximate size of the cells in FOV in pixels',
                    'timescale of sensor in deconvolution (in seconds)',
                    'sampling rate (total across planes)',
                    '0 to select num_cores, -1 to disable parallelism, N to enforce value',
                    '0 to select number of planes, -1 to disable parallelism, N to enforce value',
                    '# of subsampled frames for finding reference image',
                    'number of frames per batch',
                    'max allowed registration shift, as a fraction of frame max(width and height)',
                    'when multi-channel, you can align by non-functional channel (1-based)',
                    'max number of binned frames for the SVD',
                    'max number of SVD components to keep for ROI detection',
                    'adjust the automatically determined threshold by this scalar multiplier',
                    'minimum ratio between neuropil radius and cell radius',
                    'number of pixels between ROI and neuropil donut',
                    'maximum neuropil radius',
                    'minimum number of pixels in the neuropil',
                    'window for maximin',
                    'smoothing constant for gaussian filter',
                    'smoothing constant for gaussian filter',
                    'neuropil coefficient']
        l=0
        self.keylist = []
        self.editlist = []
        kk=0
        for lkey in keys:
            k = 0
            qlabel = QtGui.QLabel(labels[l])
            bigfont = QtGui.QFont("Arial", 10, QtGui.QFont.Bold)
            qlabel.setFont(bigfont)
            self.layout.addWidget(qlabel,0,l,1,1)
            for key in lkey:
                lops = 1
                if self.ops[key] or (self.ops[key] == 0):
                    qedit = QtGui.QLineEdit()
                    qlabel = QtGui.QLabel(key)
                    qlabel.setToolTip(tooltips[kk])
                    qedit.setText(str(self.ops[key]))
                    self.layout.addWidget(qlabel,k*2+1,l,1,1)
                    self.layout.addWidget(qedit,k*2+2,l,1,1)
                    self.keylist.append(key)
                    self.editlist.append(qedit)
                k+=1
                kk+=1
            l+=1
        btiff = QtGui.QPushButton('Add directory to data_path')
        btiff.clicked.connect(self.get_folders)
        self.layout.addWidget(btiff,0,0,1,1)
        qlabel = QtGui.QLabel('data_path')
        qlabel.setFont(bigfont)
        self.layout.addWidget(qlabel,1,0,1,1)
        btiff = QtGui.QPushButton('Choose save_path\n(default is data_path)')
        btiff.clicked.connect(self.save_folder)
        self.layout.addWidget(btiff,13,0,1,1)
        self.layout.addWidget(QtGui.QLabel('-'),14,0,1,1)
        self.runButton = QtGui.QPushButton('RUN SUITE2P')
        self.runButton.clicked.connect(lambda: self.run_S2P(parent))
        self.layout.addWidget(self.runButton,16,0,1,1)
        self.textEdit = QtGui.QTextEdit()
        self.layout.addWidget(self.textEdit, 17,0,15,l)
        self.process = QtCore.QProcess(self)
        self.process.readyReadStandardOutput.connect(self.stdout_write)
        self.process.readyReadStandardError.connect(self.stderr_write)
        # disable the button when running the s2p process
        self.process.started.connect(self.started)
        self.process.finished.connect(lambda: self.finished(parent))
        # stop process
        self.stopButton = QtGui.QPushButton('STOP')
        self.stopButton.setEnabled(False)
        self.layout.addWidget(self.stopButton, 16,1,1,1)
        self.stopButton.clicked.connect(self.stop)

    def stop(self):
        self.finish = False
        self.process.kill()

    def started(self):
        self.runButton.setEnabled(False)
        self.stopButton.setEnabled(True)

    def finished(self, parent):
        self.runButton.setEnabled(True)
        self.stopButton.setEnabled(False)
        if self.finish and not self.error:
            cursor = self.textEdit.textCursor()
            cursor.movePosition(cursor.End)
            cursor.insertText('Opening in GUI (can close this window)\n')
            parent.fname = os.path.join(self.save_path, 'suite2p', 'plane0','stat.npy')
            parent.load_proc()
        elif not self.error:
            cursor = self.textEdit.textCursor()
            cursor.movePosition(cursor.End)
            cursor.insertText('Interrupted by user (not finished)\n')
        else:
            cursor = self.textEdit.textCursor()
            cursor.movePosition(cursor.End)
            cursor.insertText('Interrupted by error (not finished)\n')

    def stdout_write(self):
        cursor = self.textEdit.textCursor()
        cursor.movePosition(cursor.End)
        cursor.insertText(str(self.process.readAllStandardOutput(), 'utf-8'))
        self.textEdit.ensureCursorVisible()

    def stderr_write(self):
        cursor = self.textEdit.textCursor()
        cursor.movePosition(cursor.End)
        cursor.insertText('>>>ERROR<<<\n')
        cursor.insertText(str(self.process.readAllStandardError(), 'utf-8'))
        self.textEdit.ensureCursorVisible()
        self.error = True

    def run_S2P(self, parent):
        self.finish = True
        self.error = False
        k=0
        for key in self.keylist:
            if type(self.ops[key]) is float:
                self.ops[key] = float(self.editlist[k].text())
                #print(key,'\t\t', float(self.editlist[k].text()))
            elif type(self.ops[key]) is int:
                self.ops[key] = int(self.editlist[k].text())
                #print(key,'\t\t', int(self.editlist[k].text()))
            elif type(self.ops[key]) is bool:
                self.ops[key] = bool(self.editlist[k].text())
                #print(key,'\t\t', bool(self.editlist[k].text()))
            k+=1
        self.db = {}
        self.db['data_path'] = self.data_path
        self.db['subfolders'] = []
        if len(self.save_path)==0:
            fpath = self.db['data_path'][0]
            self.save_path = fpath
        self.db['save_path0'] = self.save_path
        print('Running suite2p!')
        print('starting process')
        np.save('ops.npy', self.ops)
        np.save('db.npy', self.db)
        self.process.start('python -u -W ignore -m suite2p --ops ops.npy --db db.npy')

    def get_folders(self):
        name = QtGui.QFileDialog.getExistingDirectory(self, "Add directory to data path")
        self.data_path.append(name)
        self.layout.addWidget(QtGui.QLabel(name),
                              len(self.data_path)+1,0,1,1)

    def save_folder(self):
        name = QtGui.QFileDialog.getExistingDirectory(self, "Save folder for data")
        self.save_path = name
        self.layout.addWidget(QtGui.QLabel(name),14,0,1,1)

### custom QDialog which makes a list of items you can include/exclude
class ListChooser(QtGui.QDialog):
    def __init__(self, Text, parent=None):
        super(ListChooser, self).__init__(parent)
        self.setGeometry(300,300,300,320)
        self.setWindowTitle(Text)
        self.win = QtGui.QWidget(self)
        layout = QtGui.QGridLayout()
        self.win.setLayout(layout)
        #self.setCentralWidget(self.win)
        loadcell = QtGui.QPushButton('Load iscell.npy')
        loadcell.resize(200,50)
        loadcell.clicked.connect(self.load_cell)
        layout.addWidget(loadcell,0,0,1,1)
        loadtext = QtGui.QPushButton('Load txt file')
        loadcell.resize(200,50)
        #loadtext.resize(loadtext.minimumSizeHint())
        loadtext.clicked.connect(self.load_text)
        layout.addWidget(loadtext,0,1,1,1)
        self.list = QtGui.QListWidget(parent)
        layout.addWidget(self.list,1,0,5,2)
        #self.list.resize(450,250)
        self.list.setSelectionMode(QtGui.QAbstractItemView.MultiSelection)
        done = QtGui.QPushButton('OK')
        done.resize(done.minimumSizeHint())
        done.clicked.connect(lambda: self.exit_list(parent))
        layout.addWidget(done,7,0,1,2)

    def load_cell(self):
        name = QtGui.QFileDialog.getOpenFileName(self, 'Open iscell.npy file',filter='iscell.npy')
        if name:
            try:
                iscell = np.load(name[0])
                badfile = True
                if iscell.shape[0] > 0:
                    if iscell[0]==0 or iscell[0]==1:
                        badfile = False
                        self.list.addItem(name[0])

                if badfile:
                    QtGui.QMessageBox.information(self, 'iscell.npy should be 0/1')
            except (OSError, RuntimeError, TypeError, NameError):
                QtGui.QMessageBox.information(self, 'iscell.npy should be 0/1')
        else:
            QtGui.QMessageBox.information(self, 'iscell.npy should be 0/1')


    def load_text(self):
        name = QtGui.QFileDialog.getOpenFileName(self, 'Open *.txt file', filter='text file (*.txt)')
        if name:
            try:
                txtfile = open(name[0], 'r')
                files = txtfile.read()
                txtfile.close()
                files = files.splitlines()
                for f in files:
                    self.list.addItem(f)
            except (OSError, RuntimeError, TypeError, NameError):
                QtGui.QMessageBox.information(self, 'not a text file')
                print('not a good list')

    def exit_list(self, parent):
        parent.trainfiles = []
        for item in self.list.selectedItems():
            parent.trainfiles.append(item.text())
        self.accept()

### custom QPushButton class that plots image when clicked
# requires buttons to put into a QButtonGroup (parent.viewbtns)
# allows only 1 button to pressed at a time
class ViewButton(QtGui.QPushButton):
    def __init__(self, bid, Text, parent=None):
        super(ViewButton,self).__init__(parent)
        self.setText(Text)
        self.setCheckable(True)
        self.resize(self.minimumSizeHint())
        self.clicked.connect(lambda: self.press(parent, bid))
        self.show()
    def press(self, parent, bid):
        ischecked  = self.isChecked()
        if ischecked:
            parent.ops_plot[1] = bid
            M = fig.draw_masks(parent)
            parent.plot_masks(M)

### Changes colors of ROIs
# button group is exclusive (at least one color is always chosen)
class ColorButton(QtGui.QPushButton):
    def __init__(self, bid, Text, parent=None):
        super(ColorButton,self).__init__(parent)
        self.setText(Text)
        self.setCheckable(True)
        self.resize(self.minimumSizeHint())
        self.clicked.connect(lambda: self.press(parent, bid))
        self.show()
    def press(self, parent, bid):
        ischecked  = self.isChecked()
        if ischecked:
            parent.ops_plot[2] = bid
            M = fig.draw_masks(parent)
            parent.plot_masks(M)
            parent.plot_colorbar(bid)



import sys
import numpy as np
from scipy.ndimage import gaussian_filter
import os

class Classifier:
    def __init__(self, classfile=None, trainfiles=None, statclass=None):
        # load previously trained classifier
        if classfile is not None:
            self.classfile = classfile
            self.load_classifier()

        elif trainfiles is not None and statclass is not None:
            self.trainfiles = trainfiles
            self.statclass = statclass
            self.load_data()
            if self.traindata.shape[0]==0:
                self.loaded = False
            else:
                self.loaded = True
                self.train()

    def train(self):
        '''input: matrix ncells x cols, where first column are labels, and the other
                    columns are the statistics to use for classification
            output: distribution of cell likelihood for each statistic
        '''
        iscell = self.traindata[:,0].astype(bool)
        notcell = ~iscell
        stats = self.traindata[:,1:]
        # make grid of statistics values
        ncells,nstats = stats.shape
        grid = np.zeros((100, stats.shape[1]), np.float32)

        for n in range(nstats):
            grid[:,n] = np.linspace(np.percentile(stats[:,n], 2),
                                    np.percentile(stats[:,n], 98),
                                    100)
        hists = np.zeros((99,nstats,2), np.float32)
        for k in range(2):
            if k==1:
                ix = iscell
            else:
                ix = notcell
            for n in range(nstats):
                hists[:,n,k] = smooth_distribution(stats[ix,n], grid[:,n])

        self.hists = hists
        self.grid = grid

    def apply(self, stats, classval):
        '''inputs: model (from train), statistics of cells to classify, and
                    classval (probability of cell cutoff)
            output: iscell labels
        '''
        ncells, nstats = stats.shape
        grid = self.grid
        hists = self.hists
        logp = np.zeros((ncells,2), np.float32)
        for n in range(nstats):
            x = stats[:,n]
            x[x<grid[0,n]]   = grid[0,n]
            x[x>grid[-1,n]]  = grid[-1,n]
            ibin = np.digitize(x, grid[:,n], right=True) - 1
            logp = logp + np.log(np.squeeze(hists[ibin,n,:])+1e-5)
        p = np.ones((1,2),np.float32)
        p = p / p.sum()
        for n in range(10):
            L = logp + np.log(p)
            L = L - np.expand_dims(L.max(axis=1), axis=1)
            rs = np.exp(L) + 1e-5
            rs = rs / np.expand_dims(rs.sum(axis=1), axis=1)
            p = rs.mean(axis=0)

        probcell = rs[:,0]
        iscell = probcell > classval
        return iscell, probcell


    def load_classifier(self):
        try:
            model = np.load(self.classfile)
            model = model.item()
            self.grid = model['grid']
            self.hists = model['hists']
            self.trainfiles = model['trainfiles']
            self.statclass = model['statclass']
            self.loaded = True
        except (ValueError, KeyError, OSError, RuntimeError, TypeError, NameError):
            print('ERROR: incorrect classifier file')
            self.loaded = False


    def save_classifier(self, fname):
        model = {}
        model['grid'] = self.grid
        model['hists'] = self.hists
        model['trainfiles'] = self.trainfiles
        model['statclass'] = self.statclass
        print('saving classifier in ' + fname)
        np.save(fname, model)

    def load_data(self):
        statclass = self.statclass
        trainfiles = self.trainfiles
        traindata = np.zeros((0,len(statclass)+1),np.float32)
        trainfiles_good = []
        if trainfiles is not None:
            for fname in trainfiles:
                badfile = False
                basename, bname = os.path.split(fname)
                try:
                    iscell = np.load(fname)
                    ncells = iscell.shape[0]
                except (ValueError, OSError, RuntimeError, TypeError, NameError):
                    print('\t'+fname+': not a numpy array of booleans')
                    badfile = True
                if not badfile:
                    basename, bname = os.path.split(fname)
                    lstat = 0
                    try:
                        stat = np.load(basename+'/stat.npy')
                        stat = stat.item()
                        ypix = stat[0]['ypix']
                        lstat = len(stat)
                    except (KeyError, OSError, RuntimeError, TypeError, NameError):
                        print('\t'+basename+': incorrect or missing stat.npy file :(')
                    if lstat != ncells:
                        print('\t'+basename+': stat.npy is not the same length as iscell.npy')
                    else:
                        # add iscell and stat to classifier
                        print('\t'+fname+' was added to classifier')
                        iscell = iscell.astype(np.float32)
                        nall = np.zeros((ncells, len(statclass)+1),np.float32)
                        nall[:,0] = iscell
                        k=0
                        for key in statclass:
                            k+=1
                            for n in range(0,ncells):
                                nall[n,k] = stat[n][key]
                        traindata = np.concatenate((traindata,nall),axis=0)
                        trainfiles_good.append(fname)
        self.traindata = traindata
        self.trainfiles = trainfiles

def smooth_distribution(x, grid):
    xbin = x
    sig = 10.0
    xbin[xbin<grid[0]] = grid[0]#*np.ones(((xbin<grid[0]).sum(),))
    xbin[xbin>grid[-1]] = grid[-1]#*np.ones(((xbin>grid[-1]).sum(),))
    hist0 = np.histogram(xbin, grid)
    hist0 = hist0[0]
    hist = hist0#%gaussian_filter(hist0, sig)
    print(hist)
    hist = hist / hist.sum()
    return hist0



import numpy as np
import time, os
from suite2p import register, dcnv, celldetect
from scipy import stats
from multiprocessing import Pool

def tic():
    return time.time()
def toc(i0):
    return time.time() - i0

def default_ops():
    ops = {
        'save_path0': [], # default is the first item in data_path
        'diameter':12, # this is the main parameter for cell detection
        'tau':  1., # this is the main parameter for deconvolution
        'fs': 10.,  # sampling rate (total across planes)
        'nplanes' : 1, # each tiff has these many planes in sequence
        'nchannels' : 1, # each tiff has these many channels per plane
        'functional_chan' : 1, # this channel is used to extract functional ROIs (1-based)
        'align_by_chan' : 1, # when multi-channel, you can align by non-functional channel (1-based)
        'look_one_level_down': False, # whether to look in all subfolders when searching for tiffs
        'baseline': 'maximin', # baselining mode
        'win_baseline': 60., # window for maximin
        'sig_baseline': 10., # smoothing constant for gaussian filter
        'prctile_baseline': 8.,# smoothing constant for gaussian filter
        'neucoeff': .7,  # neuropil coefficient
        'neumax': 1.,  # maximum neuropil coefficient (not implemented)
        'niterneu': 5, # number of iterations when the neuropil coefficient is estimated (not implemented)
        'maxregshift': 0.1, # max allowed registration shift, as a fraction of frame max(width and height)
        'reg_tif': False, # whether to save registered tiffs for manual inspection
        'subpixel' : 10, # precision of subpixel registration (1/subpixel steps)
        'batch_size': 200, # number of frames per batch
        'num_workers': 0, # 0 to select num_cores, -1 to disable parallelism, N to enforce value
        'num_workers_roi': -1, # 0 to select number of planes, -1 to disable parallelism, N to enforce value
        'nimg_init': 200, # subsampled frames for finding reference image
        'navg_frames_svd': 5000, # max number of binned frames for the SVD
        'nsvd_for_roi': 1000, # max number of SVD components to keep for ROI detection
        'max_iterations': 10, # maximum number of iterations to do cell detection
        'ratio_neuropil': 3., # minimum ratio between neuropil radius and cell radius
        'tile_factor': 1, # use finer (>1) or coarser (<1) tiles for neuropil estimation
        'threshold_scaling': 1, # adjust the automatically determined threshold by this scalar multiplier        
        'inner_neuropil_radius': 2, # number of pixels to keep between ROI and neuropil donut
        'outer_neuropil_radius': np.inf, # maximum neuropil radius
        'min_neuropil_pixels': 350, # minimum number of pixels in the neuropil
        'ratio_neuropil_to_cell': 3, # minimum ratio between neuropil radius and cell radius
        'allow_overlap': False,                 
      }
    return ops

def get_cells(ops):
    i0 = tic()
    ops, stat = celldetect.sourcery(ops)
    print('time %4.4f. Found %d ROIs'%(toc(i0), len(stat)))
    # extract fluorescence and neuropil
    F, Fneu = celldetect.extractF(ops, stat)
    print('time %4.4f. Extracted fluorescence from %d ROIs'%(toc(i0), len(stat)))
    # subtract neuropil
    dF = F - ops['neucoeff'] * Fneu
    # deconvolve fluorescence
    spks = dcnv.oasis(dF, ops)
    print('time %4.4f. Detected spikes in %d ROIs'%(toc(i0), len(stat)))
    # compute activity statistics for classifier
    sk = stats.skew(dF, axis=1)
    for k in range(F.shape[0]):
        stat[k]['skew'] = sk[k]
    # save results
    np.save(ops['ops_path'], ops)
    fpath = ops['save_path']
    np.save(os.path.join(fpath,'F.npy'), F)
    np.save(os.path.join(fpath,'Fneu.npy'), Fneu)
    np.save(os.path.join(fpath,'spks.npy'), spks)
    np.save(os.path.join(fpath,'stat.npy'), stat)
    print('results saved to %s'%ops['save_path'])

    return ops

def run_s2p(ops={},db={}):
    i0 = tic()

    ops = {**ops, **db}

    if 'save_path0' not in ops or len(ops['save_path0'])==0:
        ops['save_path0'] = ops['data_path'][0]

    # check if there are files already registered
    fpathops1 = os.path.join(ops['save_path0'], 'suite2p', 'ops1.npy')
    files_found_flag = True
    if os.path.isfile(fpathops1):
        ops1 = np.load(fpathops1)
        files_found_flag = True
        for i,op in enumerate(ops1):
            files_found_flag &= os.path.isfile(op['reg_file'])
            # use the new options
            ops1[i] = {**op, **ops}
    else:
        files_found_flag = False

    if not files_found_flag:
        # get default options
        ops0 = default_ops()
        # combine with user options
        ops = {**ops0, **ops}
        # copy tiff to a binary
        ops1 = register.tiff_to_binary(ops)
        print('time %4.4f. Wrote tifs to binaries for %d planes'%(toc(i0), len(ops1)))
        # register tiff
        ops1 = register.register_binary(ops1)
        # save ops1
        np.save(fpathops1, ops1)
        print('time %4.4f. Registration complete'%toc(i0))
    else:
        print('found ops1 and pre-registered binaries')
        print(ops1[0]['reg_file'])
        print('overwriting ops1 with new ops')
        print('skipping registration...')
        
    if len(ops1)>1 and ops['num_workers_roi']>=0:
        if ops['num_workers_roi']==0:
            ops['num_workers_roi'] = len(ops1)
        with Pool(ops['num_workers_roi']) as p:
            results = p.map(get_cells, ops1)
        for k in range(len(ops1)):
            ops1[k] = results[k]
    else:
        for k in range(len(ops1)):
            ops1[k] = get_cells(ops1[k])

    # save final ops1 with all planes
    np.save(fpathops1, ops1)

    print('finished all tasks in %4.4f sec'%toc(i0))

    return ops1
